/*********************************************************************
 *
 *	Address Resolution Protocol (ARP) Client and Server
 *  Module for Microchip TCP/IP Stack
 *	 -Provides IP address to Ethernet MAC address translation
 *	 -Reference: RFC 826
 *
 *********************************************************************
 * FileName:        ARP.c
 * Dependencies:    string.h
 *                  StackTsk.h
 *                  Helpers.h
 *                  ARP.h
 *                  MAC.h
 * Processor:       PIC18, PIC24F, PIC24H, dsPIC30F, dsPIC33F, PIC32
 * Compiler:        Microchip C32 v1.00 or higher
 *					Microchip C30 v3.01 or higher
 *					Microchip C18 v3.13 or higher
 *					HI-TECH PICC-18 STD 9.50PL3 or higher
 * Company:         Microchip Technology, Inc.
 *
 * Software License Agreement
 *
 * Copyright (C) 2002-2008 Microchip Technology Inc.  All rights 
 * reserved.
 *
 * Microchip licenses to you the right to use, modify, copy, and 
 * distribute: 
 * (i)  the Software when embedded on a Microchip microcontroller or 
 *      digital signal controller product ("Device") which is 
 *      integrated into Licensee's product; or
 * (ii) ONLY the Software driver source files ENC28J60.c and 
 *      ENC28J60.h ported to a non-Microchip device used in 
 *      conjunction with a Microchip ethernet controller for the 
 *      sole purpose of interfacing with the ethernet controller. 
 *
 * You should refer to the license agreement accompanying this 
 * Software for additional information regarding your rights and 
 * obligations.
 *
 * THE SOFTWARE AND DOCUMENTATION ARE PROVIDED "AS IS" WITHOUT 
 * WARRANTY OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT 
 * LIMITATION, ANY WARRANTY OF MERCHANTABILITY, FITNESS FOR A 
 * PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT SHALL 
 * MICROCHIP BE LIABLE FOR ANY INCIDENTAL, SPECIAL, INDIRECT OR 
 * CONSEQUENTIAL DAMAGES, LOST PROFITS OR LOST DATA, COST OF 
 * PROCUREMENT OF SUBSTITUTE GOODS, TECHNOLOGY OR SERVICES, ANY CLAIMS 
 * BY THIRD PARTIES (INCLUDING BUT NOT LIMITED TO ANY DEFENSE 
 * THEREOF), ANY CLAIMS FOR INDEMNITY OR CONTRIBUTION, OR OTHER 
 * SIMILAR COSTS, WHETHER ASSERTED ON THE BASIS OF CONTRACT, TORT 
 * (INCLUDING NEGLIGENCE), BREACH OF WARRANTY, OR OTHERWISE.
 *
 *
 * Author               Date    Comment
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * Nilesh Rajbharti     5/1/01  Original        (Rev 1.0)
 * Nilesh Rajbharti     2/9/02  Cleanup
 * Nilesh Rajbharti     5/22/02 Rev 2.0 (See version.log for detail)
 * Howard Schlunder		8/17/06	Combined ARP.c and ARPTsk.c into ARP.c; 
 *								rewrote some of it to look more linear
 ********************************************************************/
#define __ARP_C

#include "TCPIP Stack/TCPIP.h"

#include "router.h"


/****************************************************************************
  Section:
	Constants and Variables
  ***************************************************************************/


#define ARP_OPERATION_REQ       0x01u		// Operation code indicating an ARP Request
#define ARP_OPERATION_RESP      0x02u		// Operation code indicating an ARP Response

#define HW_ETHERNET             (0x0001u)	// ARP Hardware type as defined by IEEE 802.3
#define ARP_IP                  (0x0800u)	// ARP IP packet type as defined by IEEE 802.3


#pragma udata arp_table
struct ARP_TABLE arp_table;
#pragma code


// ARP packet structure
typedef struct __attribute__((aligned(2), packed))
{
    WORD        HardwareType;
    WORD        Protocol;
    BYTE        MACAddrLen;
    BYTE        ProtocolLen;
    WORD        Operation;
    MAC_ADDR    SenderMACAddr;
    IP_ADDR     SenderIPAddr;
    MAC_ADDR    TargetMACAddr;
    IP_ADDR     TargetIPAddr;
} ARP_PACKET;



/****************************************************************************
  Section:
	Helper Function Prototypes
  ***************************************************************************/

static void SwapARPPacket(ARP_PACKET* p);
static BOOL ARPPut(BYTE CSX, ARP_PACKET* packet);



/****************************************************************************
  Section:
	Function Implementations
  ***************************************************************************/



void arp_add_entry(BYTE CSX, IP_ADDR *ip, MAC_ADDR *mac)
  {
    int i = 0;

    i = ARPIsResolved(CSX, ip, NULL );

    if (i)
      i--;
    else
      {
        i = arp_table.pointer[CSX]++;
        arp_table.pointer[CSX] = arp_table.pointer[CSX] % ARP_TABLE_LENGTH;
      }

    arp_table.csx[CSX].entry[i].IPAddr = *ip;
    arp_table.csx[CSX].entry[i].MACAddr = *mac;
  }


/*****************************************************************************
  Function:
	static BOOL ARPPut(ARP_PACKET* packet)

  Description:
	Writes an ARP packet to the MAC.

  Precondition:
	None

  Parameters:
	packet - A pointer to an ARP_PACKET structure with correct operation
				and target preconfigured.

  Return Values:
  	TRUE - The ARP packet was generated properly
  	FALSE - Not a possible return value
  ***************************************************************************/
static BOOL ARPPut(BYTE CSX, ARP_PACKET* packet)
{ 
	while(!MACIsTxReady(CSX));
	MACSetWritePtr(CSX,BASE_TX_ADDR);
	

    packet->HardwareType  = HW_ETHERNET;
    packet->Protocol      = ARP_IP;
    packet->MACAddrLen    = sizeof(MAC_ADDR);
    packet->ProtocolLen   = sizeof(IP_ADDR);

    SwapARPPacket(packet);

    MACPutHeader(CSX, &packet->TargetMACAddr, INTF_MAC_ADDR(CSX), MAC_ARP, sizeof(*packet));
    MACPutArray(CSX,(BYTE*)packet, sizeof(*packet));
    MACFlush(CSX);
	
	return TRUE;
}



/*****************************************************************************
  Function:
	void ARPInit(void)

  Summary:
	Initializes the ARP module.
	
  Description:
  	Initializes the ARP module.  Call this function once at boot to 
  	invalidate the cached lookup.

  Precondition:
	None

  Parameters:
	None

  Returns:
  	None
  
  Remarks:
  	This function is only required when the stack is a client, and therefore
  	is only enabled when STACK_CLIENT_MODE is enabled.
  ***************************************************************************/

void ARPInit(void)
  {
    BYTE i,j;

    for (j=0u; j<NOI; j++)
      {
        arp_table.pointer[j]=0;
        for (i=0u; i<ARP_TABLE_LENGTH; i++)
          {
            memset((void*) &(arp_table.csx[j].entry[i].MACAddr),0xFF,sizeof(MAC_ADDR));
            arp_table.csx[j].entry[i].IPAddr.Val = 0ul;
          }
      }
  }


static BOOL proxy_arp(BYTE CSX, ARP_PACKET *packet)
  {
    BYTE i;
    
    int from_local_network = -1;  // not from local network
    int to_local_network   = -1;  // not to local network
    
    // if the packet came from a network for which we are a specific route 
    // and it is for something not on that network then respond with interface number
    
    // if the packet came from a netowkr with a default route and it is 
    // for something on our netowrk then return with the interface number.

    // NOTE: Hmmm maybe I should index all of this information!  
    
    // iterate through the interfaces
    for (i=0u; i<number_of_routes; i++)
      {
        // skip interfaces with a default route
        if (route[i].mask.Val)
          {
            if ((packet->SenderIPAddr.Val & route[i].mask.Val) == route[i].net.Val)
              {
                // the sender is from an IP address indicating a local network.
                from_local_network = route[i].ino;
                break;
              }
          }
      } // for
      
    for (i=0u; i<number_of_routes; i++)
      {
        // skip interfaces with a default route
        if (route[i].mask.Val)
          { 
            if ((packet->TargetIPAddr.Val & route[i].mask.Val) == route[i].net.Val)
              {
                to_local_network = route[i].ino;
                break;
              }
          }
      } // for
      
      // if it is from a non-local net and to a non-local net then don't proxy              
      // if it is from a non-local net and to a local net then do proxy
            
      // if it is from a local net to a non-local net then do proxy
      // if it is from the same two local nets then don't proxy.
      // if it is from two local nets that are not the same then do proxy.
      
      if (from_local_network < 0)
        {
          // it is from a non-local network
          
          if (to_local_network < 0) 
            return FALSE;    // and to a non-local network.
          else
            return TRUE;     // it is from a non-local network and to a local network.
        }
      else
        {  // it is from a local network
          
          if (to_local_network < 0) 
            return TRUE;    // and to a non-local network.
          else
            {  // it is from a local network and to a local network.
              if (to_local_network == from_local_network)
                return FALSE; // it is the same local network 
              else
                return TRUE; // two different local networks.
            }
        }
        
    // this should never happen.
    return FALSE;
  }
  
/*****************************************************************************
  Function:
	BOOL ARPProcess(void)

  Summary:
	Processes an incoming ARP packet.
	
  Description:
  	Retrieves an ARP packet from the MAC buffer and determines if it is a
  	response to our request (in which case the ARP is resolved) or if it
  	is a request requiring our response (in which case we transmit one.)

  Precondition:
	ARP packet is ready in the MAC buffer.

  Parameters:
	None

  Return Values:
  	TRUE - All processing of this ARP packet is complete.  Do not call 
  			again until a new ARP packet is waiting in the RX buffer.
  	FALSE - This function must be called again.  More time is needed to 
  			send an ARP response.
  ***************************************************************************/
BOOL ARPProcess(BYTE CSX)
{  
	ARP_PACKET packet;
	static NODE_INFO Target;
	static enum
	{
	    SM_ARP_IDLE = 0,
	    SM_ARP_REPLY
	} smARP = SM_ARP_IDLE;

    switch(smARP)
    {
	    case SM_ARP_IDLE:
			// Obtain the incoming ARP packet
		    MACGetArray(CSX,(BYTE*)&packet, sizeof(packet));		
		    MACDiscardRx(CSX);
		    SwapARPPacket(&packet);
		
			// Validate the ARP packet
		    if ( packet.HardwareType != HW_ETHERNET     ||
		         packet.MACAddrLen != sizeof(MAC_ADDR)  ||
		         packet.ProtocolLen != sizeof(IP_ADDR) )
		    {
		         return TRUE;
		    }
		
			// Handle incoming ARP responses
			if(packet.Operation == ARP_OPERATION_RESP)
			{
  			// was this sent to us?  MAC_CMP returns 0 if equal
  			if (MAC_CMP(&packet.TargetMACAddr,&intf[CSX].mac))
  			  return TRUE;

        /* RFC1812
         * A router MUST not believe any ARP reply that claims that the Link
         * Layer address of another host or router is a broadcast or multicast
         * address.
         */

        if (IS_BCAST_1(packet.SenderIPAddr.Val)) return TRUE;
        if (IS_BCAST_0(packet.SenderIPAddr.Val)) return TRUE;
        if (IS_LOOPBACK(packet.SenderIPAddr.Val)) return TRUE; 
        if (IS_LTD_BCAST_1(packet.SenderIPAddr.Val,intf[CSX].ip.Val,intf[CSX].mask.Val)) return TRUE;
        if (IS_LTD_BCAST_0(packet.SenderIPAddr.Val,intf[CSX].ip.Val,intf[CSX].mask.Val)) return TRUE;
        if (IS_MCAST(packet.SenderIPAddr.Val)) return TRUE;
			  
  			// This response was sent to us.  Save the entry.
  			arp_add_entry(CSX,&packet.SenderIPAddr, &packet.SenderMACAddr);
			
				return TRUE;
			}

			// Handle incoming ARP requests for our MAC address
			if(packet.Operation == ARP_OPERATION_REQ)
			{
  			/* Is the arp request for this interface? */
				if(packet.TargetIPAddr.Val != intf[CSX].ip.Val)
					{
  					// it is not for this interface, but I still might need to respond 
  				  if (! proxy_arp(CSX,&packet))
  				    { 
    				    // we will not proxy arp this requst.
    				    // return true to indicate we are finished with this packet.
    				    return TRUE;
    				  }
          }
				
				// save the original sender's information
				Target.IPAddr = packet.SenderIPAddr;
				Target.MACAddr = packet.SenderMACAddr;
				
				// the old target will now be the "sender"
				packet.SenderIPAddr  = packet.TargetIPAddr;

        // add our mac
	      memcpy(&packet.SenderMACAddr, (void*)INTF_MAC_ADDR(CSX), sizeof(packet.SenderMACAddr));
	  
	      // copy in the new target information
        packet.TargetMACAddr	= Target.MACAddr;
    	  packet.TargetIPAddr		= Target.IPAddr;
    
				smARP = SM_ARP_REPLY;
			}
			// Do not break.  If we get down here, we need to send a reply.	

	    case SM_ARP_REPLY:
	        packet.Operation		= ARP_OPERATION_RESP;

    			// Send an ARP response to a previously received request
    			if(!ARPPut(CSX,&packet))
    			{
    	           return FALSE;
    			}
    
    			// Begin listening for ARP requests again
    			smARP = SM_ARP_IDLE;
	        break;
	}

    return TRUE;
}



/*****************************************************************************
  Function:
	void ARPResolve(IP_ADDR* IPAddr)

  Summary:
	Transmits an ARP request to resolve an IP address.
	
  Description:
  	This function transmits and ARP request to determine the hardware
  	address of a given IP address.

  Precondition:
	ARP packet is ready in the MAC buffer.

  Parameters:
	IPAddr - The IP address to be resolved

  Returns:
  	None

  Remarks:
  	This function is only required when the stack is a client, and therefore
  	is only enabled when STACK_CLIENT_MODE is enabled.

  MIKE: The calling routine should determine if this is the apropriate interface for
  this type of arp request.  This function only sends the request.  If does not verify
  whether the request is apropriate.
  There should be a delay to prevent arp flooding the network.
  ***************************************************************************/

void ARPResolve(BYTE CSX, IP_ADDR* IPAddr)
{
  ARP_PACKET packet;
  
  // NOTE: Should this be a limited broadcast?
  
	packet.Operation            = ARP_OPERATION_REQ;
	packet.TargetMACAddr.v[0]   = 0xff;
	packet.TargetMACAddr.v[1]   = 0xff;
	packet.TargetMACAddr.v[2]   = 0xff;
	packet.TargetMACAddr.v[3]   = 0xff;
	packet.TargetMACAddr.v[4]   = 0xff;
	packet.TargetMACAddr.v[5]   = 0xff;
 
	packet.TargetIPAddr			=  *IPAddr;

				// the old target will now be the "sender"
	packet.SenderIPAddr  = intf[CSX].ip;

        // add our mac
	memcpy(&packet.SenderMACAddr, (void*)INTF_MAC_ADDR(CSX), sizeof(packet.SenderMACAddr));
	      
	ARPPut(CSX,&packet);
}




/*****************************************************************************
  Function:
	BOOL ARPIsResolved(IP_ADDR* IPAddr, MAC_ADDR* MACAddr)

  Summary:
	Determines if an ARP request has been resolved yet.
	
  Description:
  	This function checks if an ARP request has been resolved yet, and if
  	so, stores the resolved MAC address in the pointer provided.

  Precondition:
	ARP packet is ready in the MAC buffer.

  Parameters:
	IPAddr - The IP address to be resolved
	MACAddr - A buffer to store the corresponding MAC address

  Return Values:
  	TRUE - The IP address has been resolved and MAC address indicates the
  			response.
  	FALSE -	The IP address is not resolved.

  Remarks:
  	This function is only required when the stack is a client, and therefore
  	is only enabled when STACK_CLIENT_MODE is enabled.
  ***************************************************************************/

BYTE ARPIsResolved(BYTE CSX, IP_ADDR *ip, MAC_ADDR *mac)
  {
    BYTE i;
    for (i=0u; i<ARP_TABLE_LENGTH; i++)
      {
        if (arp_table.csx[CSX].entry[i].IPAddr.Val == ip->Val)
          {
            if (mac)
              *mac = arp_table.csx[CSX].entry[i].MACAddr;
            return (1u+i); 
          }
      }
    return 0;
  }




/*****************************************************************************
  Function:
	static void SwapARPPacket(ARP_PACKET* p)

  Description:
  	Swaps endian-ness of header information in an ARP packet.

  Precondition:
	None

  Parameters:
	p - The ARP packet to be swapped

  Returns:
  	None
  ***************************************************************************/
static void SwapARPPacket(ARP_PACKET* p)
{
    p->HardwareType     = swaps(p->HardwareType);
    p->Protocol         = swaps(p->Protocol);
    p->Operation        = swaps(p->Operation);
}

