#!/usr/bin/perl

# This is a very simple parser for a router (and someday a firewall) configuration

use strict;

use FW::Protocol;
use FW::IP;
use FW::Port;
use FW::Action;

use FW::Interface;
use FW::Route;
use FW::Syslog;

my $tree = [];

open(IN,"<rules.txt") || die "Cannot open rules file";

our @rules;
our @routes;
our @intrfs;
our $syslog;


# allocate interface objects
for (my $i = 0; $i < 3; $i++)
  {
    push(@intrfs, new FW::Interface($i));
  }


while (my $line = <IN>)
  {
    $line =~ s/#.*$//; # remove comments
    $line =~ s/^\s+//; # remove space at beginning
    $line =~ s/\s+$//; # remove space at ending
    $line =~ s/\s+/ /g; # merge multiple spaces to a single space 

    next unless ($line);

    if ($line =~ /^syslog /)
      {
        # syslog 192.168.3.4
        my ($value) = ($line =~ /syslog (\S+)/);
        next unless ($value);

        $syslog = new FW::Syslog($value);
        print "syslog $value\n";
      }

    if ($line =~ /^interface /)
      {
        # interface 0 mac address 00:00:1d:08:42:ac
        # interface 0 ip address 192.168.2.10/24
        my ($number,$type,$value) = ($line =~ /interface (\d+) (mac|ip) address (\S+)/);
        next unless ($value);

        print "interface ($number,$type,$value)\n";
        if ($type eq 'mac')
          {
            # NOTE: check to make sure $value is indeed a MAC address
            $intrfs[$number]->setMAC($value);
          }

        if ($type eq 'ip')
          {
            # NOTE: check to make sure $value is indeed an IP address
            $intrfs[$number]->setIP($value);
          }
      }

    if ($line =~ /^route /)
      {
        # route 192.168.2.0/24 192.168.2.1
        # route default 192.168.2.1
        # route default interface 0
        $line =~ s/ default / 0.0.0.0\/0 /;

        my $route = new FW::Route();
        if ($line =~ / interface /)
          {
            my ($ip,$int) = ($line =~ /route (\d+\.\d+\.\d+\.\d+\/\d+) interface (\d+)/);
            $route->setNetwork($ip);
            $route->setInterface($int);
          }
        else
          {
            my ($net,$int) = ($line =~ /route (\d+\.\d+\.\d+\.\d+\/\d+) (\d+\.\d+\.\d+\.\d+)/);
            $route->setNetwork($net);
            $route->setRemoteIP($int);
          }
        push(@routes,$route);
      }

    #
    # Rules section
    #
    if ($line =~ /(^tcp |^udp )/)
      {
        # replace "any" ip with CIDR ip notation (any: -> 0.0.0.0/0)
        $line =~ s/any:/0.0.0.0\/0:/g;

        # replace single IP with CIDR notation (10.1.2.3 -> 10.1.2.3/32)
        $line =~ s/(\d+\.\d+\.\d+\.\d+):/$1\/32:/g;
    
        print $line . "\n";
    
        # tcp 10.2.1.1:any  12.15.2.2:80 deny
        my ($proto,$src_addr,$src_port,$dst_addr,$dst_port,$action) 
               = ($line =~ /(\S+) (\d+\.\d+\.\d+\.\d+\/\d+):(\S+) (\d+\.\d+\.\d+\.\d+\/\d+):(\S+) (\S+)/);
    
        next unless ($action);
    
        my $rule = {
            'PROTO'    => new FW::Protocol($proto),
            'SRC_IP'   => new FW::IP($src_addr),
            'SRC_PORT' => new FW::Port($src_port),
            'DST_IP'   => new FW::IP($dst_addr),
            'DST_PORT' => new FW::Port($dst_port),
            'ACTION'   => new FW::Action($action),
            'TEXT'     => $line,
          };
    
        # order the parameters
        add_rule($tree,$rule);
      } # end rules section
  }

route_select_interface();
syslog_select_interface();
 

# order routes by longest subnet mask
@routes = sort sort_routes @routes;


open(HOUT,">../firmware/rules.h");
print HOUT qq|
/* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 * WARNING: This is an auto generated file.  
 * Do not modify this file.  Any modifications will be lost!
 *
 */

#ifndef _RULES_H
#define _RULES_H

#if defined __GNUC__ 
  #define IP_ADDR		DWORD_VAL
#endif
|;

open(COUT,">../firmware/rules.c");

my $cols;
my $elements;

print COUT qq|
/* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 * WARNING: This is an auto generated file.  
 * Do not modify this file.  Any modifications will be lost!
 *
 */

#include "rules.h"

#if defined(__18CXX)
  #pragma idata RULES
#endif

|;

$elements = 5; # ip + ino
print HOUT "extern unsigned char SYSLOG_SERVER[$elements];\n";
print COUT "unsigned char SYSLOG_SERVER[$elements] = { ";
  print COUT format_hex(5,($syslog->toString(), sprintf("%x",$syslog->getInterface()) ));
print COUT "};\n\n";

$cols = 22;
$elements = 1+$#rules;
print HOUT "extern unsigned char RULES[$elements];\n";
print COUT "unsigned char RULES[$elements] = { ";
  print COUT format_hex($cols,@rules);
print COUT "};\n\n";

$cols = 14;
$elements = $cols*(1+$#intrfs);
print HOUT "extern unsigned char INTERFACES[$elements];\n";
print COUT "unsigned char INTERFACES[$elements] = { ";
  print COUT format_hex($cols, array_to_string(@intrfs));
print COUT "};\n\n";
$cols = 13;
$elements = $cols*(1+$#routes);
print HOUT "extern unsigned char ROUTES[$elements];\n";
print COUT "unsigned char ROUTES[$elements] = { ";
  print COUT format_hex($cols,array_to_string(@routes));
print COUT "};\n\n";

print HOUT qq|
#endif
|;

print COUT qq|

#if defined(__18CXX)
  #pragma code
#endif
|;

close(HOUT);
close(COUT);

sub array_to_string
  {
    my @out = ();
    foreach my $obj (@_)
     { push(@out,$obj->toString()); }
    return @out;
  }

sub format_hex
  {
    my $cols = shift;
    my $count = -1;

    my @src = ();
    my $out = '';

    foreach my $line (@_)
      {
        push(@src, split(/\s+/,$line));
      }

    foreach my $value (@src)
      {
        $out .= "\n    " unless (++$count % $cols);
        $out .= "0x$value";
        $out .= ($count!=$#src)?",":"\n";
      }
    return $out;
  }


sub add_rule
  {
    my $tree = shift;
    my $rule = shift;
    
    push(@rules, split(/\s+/, $rule->{'PROTO'}->getHex()    ));
    push(@rules, split(/\s+/, $rule->{'SRC_IP'}->getHex()   ));
    push(@rules, split(/\s+/, $rule->{'SRC_PORT'}->getHex() ));
    push(@rules, split(/\s+/, $rule->{'DST_IP'}->getHex()   ));
    push(@rules, split(/\s+/, $rule->{'DST_PORT'}->getHex() ));
    push(@rules, split(/\s+/, $rule->{'ACTION'}->getHex()   ));
    return;
  }

sub sort_routes
  {
    return ($b->getBits() <=> $a->getBits());
  }

sub route_select_interface
  {
    foreach my $route (@routes)
      {
        # determine to which interface this corresponds
        # look up the remote host
        unless ($route->isInterfaceDefined())
          {
            my $remote_ip = $route->getRemoteIP();
            foreach my $intf (@intrfs)
              {
                my $net = $intf->getNet();
                my $mask = $intf->getMask();
print "Testing router host $remote_ip with interface $net/$mask\t";
                if (FW::CIDR::is_ip_in_network($remote_ip,$net,$mask))
                  {
                    $route->setInterface( $intf->getInterface() );
print "Yes... route interface is now defined\n";
                    last;
                  }
print "\n";
              }
          }
      }
  }


sub syslog_select_interface
  {
            foreach my $intf (@intrfs)
              {
                my $net = $intf->getNet();
                my $mask = $intf->getMask();
print "Testing syslog host " . $syslog->getIP() . "  with interface $net/$mask\t";
                if (FW::CIDR::is_ip_in_network($syslog->getIP(),$net,$mask))
                  {
                    $syslog->setInterface( $intf->getInterface() );
print "Yes... route interface is now defined\n";
                    last;
                  }
print "\n";
              }
  }

